import random
import socket
import string
import sys

def print_bytes(octets):
    for b in octets:
        print('{:02x} '.format(b), end = '')
    print()

def prepare_data(size):
    nb = 1000
    data = ""
    for i in range(int(size/nb)):
        data = data + string.ascii_uppercase[i]*nb
    rest = size - int(size/nb)*nb
    data = data + string.ascii_uppercase[int(size/nb)]*rest

    return data

# this function is useful when we do
#
# sock = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.IPPROTO_IP)
# ipv4hdr = build_ipv4_header(src, dest, size)
# nb = sock.sendto(b"".join([ipv4hdr, data.encode()]), (dest, 0))
def build_ipv4_header(src, dest, size):
    # first 32 bits 
    #   version (4 bits), Internet Header Length (4 bits),
    #       Differentiated Service (6 bits), 
    #       Explicit Congestion Notification (2 bits),
    #       Total Length (16 bits) 
    #   version has to be 6, choose the rest two fields 0
    ver = 4 << 28         # 32 - 4 = 28
    ihl = int(20/4) << 24 # 32 - 4 - 4 = 24
    print_bytes((ver | ihl).to_bytes(4, byteorder='big'))
    ds  = 0  << 18 # 32 - 4 - 4 - 6 = 18
    ecn = 0  << 16 # 31 - 4 - 4 - 6 - 2 = 16
    ple = 20 + size
    row1 = (ver | ihl | ds | ecn | ple).to_bytes(4, byteorder='big')
    print_bytes(row1)

    # second 32 bits
    #   Identification (16 bits), Flags (3 bits), Fragment Offset (12 bits)
    pid   = random.randrange(1<<16) << 16 # 32 - 16
    flags = 0
    foff  = 0
    row2 = (pid | flags | foff).to_bytes(4, byteorder='big')
    # print_bytes(row2)
    
    # 3rd 32 bits
    #   Time to Live (8 bits), Protocol (8 bits), Header Checksum (16 bits)
    ttl   =   64 << 24 # 32 - 8 = 24
    proto = 0x90 << 16 # 32 - 8 - 8 = 16
    checksum = 0       # request IP to compute checksum
    row3 = (ttl | proto | checksum).to_bytes(4, byteorder='big')
    # print_bytes(row3)

    row4 = socket.inet_pton(socket.AF_INET, src)
    # print_bytes(row4)

    row5 = socket.inet_pton(socket.AF_INET, dest)
    # print_bytes(row5)

    ipv4hdr = b"".join([row1, row2, row3, row4, row5])
    print_bytes(ipv4hdr)

    return ipv4hdr;
    

def transmit(src, dest, size):
    data = prepare_data(size)
    # See /usr/include/netinet/in.h. Existing protocols may not support
    # raw packets, as such, use unused one
    sock = socket.socket(socket.AF_INET, socket.SOCK_RAW, 201)
    # The Linux manual page of raw(7) states
    # "When the IP_HDRINCL option is set, datagrams will not be fragmented and
    #  are limited to the interface MTU"
    # The default value in the testing environment is unset. But replacing 0
    # with 1 below will disable IPv4 fragmentation
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 0)

    nb = sock.sendto(data.encode(), (dest, 0))
    print("Sent an IPv4 payload of " + str(nb) + " bytes")
    
    sock.close()
    
def usage(prog):
    print("Usage: " + prog + " source desintation payload_length")

def main(argv):
    if len(sys.argv) < 4:
        usage(argv[0])
        sys.exit(1)
    transmit(argv[1], argv[2], int(argv[3]))

if __name__ == "__main__":
    main(sys.argv)
